# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An :term:`Adapter` is a physical adapter card (e.g. OSA-Express adapter,
Crypto adapter) or a logical adapter (e.g. HiperSockets switch).

Adapter resources are contained in :term:`CPC` resources.

Adapter resources originally were supported only for CPCs in DPM mode. Since
z16, Adapter resources are in addition supported for CPCs in classic mode, but
in a quite limited fashion: The only HMC operations supported with classic mode
CPCs are "List Permitted Adapters" and "Update Adapter Firmware".

Each method description of adapter related zhmcclient classes details on which
types of CPCs it is supported.

A CPC in DPM mode automatically discovers the physical adapter cards that are
plugged into the system. A CPC in classic mode shows only the adapters that
have been defined in the active IOCDS.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._port import PortManager
from ._logging import logged_api_call
from ._exceptions import CeasedExistence, ConsistencyError

from ._utils import repr_dict, repr_manager, repr_timestamp, matches_filters, \
    divide_filter_args, make_query_str, RC_ADAPTER, repr_obj_id

__all__ = ['AdapterManager', 'Adapter']


class AdapterManager(BaseManager):
    """
    Manager providing access to the :term:`Adapters <Adapter>` in a particular
    :term:`CPC`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Cpc` object:

    * :attr:`~zhmcclient.Cpc.adapters`

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    def __init__(self, cpc):
        # This function should not go into the docs.
        # Parameters:
        #   cpc (:class:`~zhmcclient.Cpc`):
        #     CPC defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'name',

            # The adapter-id property is supported for filtering, but due to
            # a firmware defect, adapters with a hex digit in their adapter-id
            # property are not found. Disabling the property causes it to
            # be handled via client-side filtering, so that mitigates the
            # defect.
            # TODO: Re-enable the property once the defect is fixed and the fix
            # is rolled out broadly enough.
            # 'adapter-id',

            'adapter-family',
            'type',
            'status',
        ]

        super().__init__(
            resource_class=Adapter,
            class_name=RC_ADAPTER,
            session=cpc.manager.session,
            parent=cpc,
            base_uri='/api/adapters',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: :term:`CPC` defining the scope for this
        manager.
        """
        return self._parent

    @logged_api_call
    # pylint: disable=arguments-differ
    def list(self, full_properties=False, filter_args=None,
             additional_properties=None):
        """
        List the Adapters in this CPC.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * SE version >= 2.13.1
        * The CPC must be in DPM mode.

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to any Adapter to be included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            default properties.

            This parameter requires HMC 2.16.0 or higher.

        Returns:

          : A list of :class:`~zhmcclient.Adapter` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.FilterConversionError`
        """
        result_prop = 'adapters'
        list_uri = f'{self.cpc.uri}/adapters'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args,
            additional_properties)

    @logged_api_call
    def create_hipersocket(self, properties):
        """
        Create and configure a HiperSockets Adapter in this CPC.

        z17 CPCs removed the "Create Hipersocket" operation. This method uses
        the "Create Hipersocket" operation z15 and earlier CPCs, and the
        "Create Partition Link" operation on z16 and later CPCs. The distinction
        is made based on the "dpm-hipersockets-partition-link-management"
        API feature on the CPC.

        HMC/SE version requirements:

        * SE version >= 2.13.1
        * The CPC must be in DPM mode.

        Authorization requirements:

        * Object-access permission to the scoping CPC.
        * On CPCs of z15 and earlier: Task permission to the
          "Create HiperSockets Adapter" task.
        * On CPCs of z16 and later: Task permission to the
          "Create Partition Link" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are the following:

            * "name" (str): Required: The adapter's 'name' property.

            * "description" (str): Optional: The adapter's 'description'
              property. Default: An empty string.

            * "maximum-transmission-unit-size" (int): Optional: The adapter's
              'maximum-transmission-unit-size' property. Default: 8

            * "port-description" (str): Optional: The adapter port's
              'description' property. Default: An empty string. This property
              is ignored on CPCs of z16 and later.

            These properties are defined in section 'Request body contents'
            in section 'Create Hipersocket' in the :term:`HMC API` book, but
            they are valid for both approaches.

        Returns:

          :class:`~zhmcclient.Adapter`:
            The resource object for the new HiperSockets Adapter.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        cpc = self.cpc
        pl_feature = cpc.api_feature_enabled(
            "dpm-hipersockets-partition-link-management")
        if pl_feature:
            name = properties['name']
            console = cpc.manager.console
            pl_name = f"Hipersocket {name}"
            if len(pl_name) > 64:
                pl_name = f"HS {name}"
                if len(pl_name) > 64:
                    pl_name = pl_name[0:64]
            part_link = console.partition_links.create({
                "name": f"HS {properties['name']}",
                "type": "hipersockets",
                "cpc-uri": cpc.uri,
            })
            props = {
                "name": name
            }
            description = properties.get("description")
            if description:
                props["description"] = description
            mtu_size = properties.get("maximum-transmission-unit-size")
            if mtu_size:
                props["maximum-transmission-unit-size"] = mtu_size
            # TODO: Use "port-description" to set port description
            uri = part_link.get_property('adapter-uri')
            adapter = Adapter(self, uri, name, props)
            adapter.update_properties(props)
        else:
            result = self.session.post(self.cpc.uri + '/adapters',
                                       body=properties)
            # There should not be overlaps, but just in case there are, the
            # returned props should overwrite the input props:
            props = copy.deepcopy(properties)
            props.update(result)
            name = props.get(self._name_prop, None)
            uri = props[self._uri_prop]
            adapter = Adapter(self, uri, name, props)

        self._name_uri_cache.update(name, uri)
        return adapter


class Adapter(BaseResource):
    """
    Representation of an :term:`Adapter`.

    Note that adapter objects do not correspond 1:1 with the physical adapter
    cards. Adapter objects do correspond 1:1 with PCHIDs, though.
    On classic mode CPCs, there are additional adapter objects that correspond
    1:1 with the VCHIDs of Network Express adapters.

    Some examples:

    * OSA Express cards with 2 ports have a single PCHID for the physical card
      and thus a single Adapter object.
    * Network Express cards with 2 ports have one PCHID for each port, and thus
      2 Adapter objects. On classic mode CPCs, each port has two additional
      VCHIDs for the 'osh' and 'neth' support of the adapter, so a 2-port card
      has 4 additional Adapter objects.
    * FICON Express cards with 2 ports have one PCHID for each port, and thus
      2 Adapter objects.
    * Crypto Express cards with 2 HSMs (= "ports") have one PCHID for each HSM,
      and thus 2 Adapter objects.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    For the properties of an Adapter, see section 'Data model' in section
    'Adapter object' in the :term:`HMC API` book.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.AdapterManager`).

    HMC/SE version requirements:

    * SE version >= 2.13.1
    """

    # Name of property for port URIs, dependent on adapter family
    port_uris_prop_by_family = {
        'ficon': 'storage-port-uris',
        'osa': 'network-port-uris',
        'roce': 'network-port-uris',
        'hipersockets': 'network-port-uris',
        'cna': 'network-port-uris',
        'cloud-network': 'network-port-uris',  # for preliminary driver
        'network-express': 'network-port-uris',
        'networking': 'network-port-uris',
    }

    # URI segment for port URIs, dependent on adapter family
    port_uri_segment_by_family = {
        'ficon': 'storage-ports',
        'osa': 'network-ports',
        'roce': 'network-ports',
        'hipersockets': 'network-ports',
        'cna': 'network-ports',
        'cloud-network': 'network-ports',  # for preliminary driver
        'network-express': 'network-ports',
        'networking': 'network-ports',
    }

    # Port type, dependent on adapter family
    port_type_by_family = {
        'ficon': 'storage',
        'osa': 'network',
        'roce': 'network',
        'hipersockets': 'network',
        'cna': 'network',
        'cloud-network': 'network',  # for preliminary driver
        'network-express': 'network',
        'networking': 'network',
    }

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.AdapterManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, AdapterManager), (
            f"Adapter init: Expected manager type {AdapterManager}, got "
            f"{type(manager)}")
        super().__init__(manager, uri, name, properties)
        # The manager objects for child resources (with lazy initialization):
        self._ports = None
        self._port_uris_prop = None
        self._port_uri_segment = None

    @property
    def ports(self):
        """
        :class:`~zhmcclient.PortManager`: Access to the :term:`Ports <Port>` of
        this Adapter.
        """
        # We do here some lazy loading.
        if not self._ports:
            family = self.get_property('adapter-family')
            try:
                port_type = self.port_type_by_family[family]
            except KeyError:
                port_type = None
            self._ports = PortManager(self, port_type)
        return self._ports

    @property
    def port_uris_prop(self):
        """
        :term:`string`: Name of adapter property that specifies the adapter
        port URIs, or the empty string ('') for adapters without ports.

        For example, 'network-port-uris' for a network adapter.
        """
        if self._port_uris_prop is None:
            family = self.get_property('adapter-family')
            try:
                self._port_uris_prop = self.port_uris_prop_by_family[family]
            except KeyError:
                self._port_uris_prop = ''
        return self._port_uris_prop

    @property
    def port_uri_segment(self):
        """
        :term:`string`: Adapter type specific URI segment for adapter port
        URIs, or the empty string ('') for adapters without ports.

        For example, 'network-ports' for a network adapter.
        """
        if self._port_uri_segment is None:
            family = self.get_property('adapter-family')
            try:
                self._port_uri_segment = self.port_uri_segment_by_family[
                    family]
            except KeyError:
                self._port_uri_segment = ''
        return self._port_uri_segment

    @property
    @logged_api_call
    def maximum_crypto_domains(self):
        """
        Integer: The maximum number of crypto domains on this crypto adapter.

        HMC/SE version requirements:

        * The parent CPC of the adapter must be in DPM mode.

        The maximum number of crypto domains on any crypto adapter (= HSM)
        is always equal to the maximum number of active partitions / LPARs
        on the CPC.

        For example, high-end systems support a maximum number of 85 crypto
        domains per HSM, and mid-range systems support a maximum number of 40
        crypto domains per HSM.

        If this adapter is not a crypto adapter, `None` is returned.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`ValueError`: Unknown crypto card type
        """
        if self.get_property('adapter-family') != 'crypto':
            return None
        card_type = self.get_property('detected-card-type')
        if card_type.startswith('crypto-express-'):
            max_domains = self.manager.cpc.maximum_active_partitions
        else:
            raise ValueError(f"Unknown crypto card type: {card_type!r}")
        return max_domains

    @logged_api_call
    def delete(self):
        """
        Delete this Adapter.

        The Adapter must be a HiperSockets Adapter and must not currently be
        the backing adapter of any partition NICs.

        z17 CPCs removed the "Delete Hipersocket" operation. This method uses
        the "Delete Hipersocket" operation z15 and earlier CPCs, and the
        "Delete Partition Link" operation on z16 and later CPCs. The distinction
        is made based on the "dpm-hipersockets-partition-link-management"
        API feature on the CPC.

        HMC/SE version requirements:

        * SE version >= 2.13.1
        * The parent CPC of the adapter must be in DPM mode.

        Authorization requirements:

        * Object-access permission to the HiperSockets Adapter to be deleted.
        * On CPCs of z15 and earlier: Task permission to the
          "Delete HiperSockets Adapter" task.
        * On CPCs of z16 and later: Task permission to the
          "Delete Partition Link" task.

        Raises:

          :exc:`~zhmcclient.CeasedExistence`: The Partition Link for the
            adapter no longer exists (on CPCs of z16 and later)
          :exc:`~zhmcclient.ConsistencyError`: Found more than one Partition
            Link for the adapter (on CPCs of z16 and later)
          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        cpc = self.manager.cpc
        pl_feature = cpc.api_feature_enabled(
            "dpm-hipersockets-partition-link-management")
        if pl_feature:
            console = cpc.manager.console
            filter_args = {"cpc-uri": cpc.uri}
            part_links = console.partition_links.list(
                filter_args=filter_args,
                additional_properties=['adapter-uri'],
            )
            part_links_for_adapter = []
            for part_link in part_links:
                if part_link.get_property('adapter-uri') == self.uri:
                    part_links_for_adapter.append(part_link)
            num_pl = len(part_links_for_adapter)
            # On z16 or later CPCs, the HMC ensures that Hipersocket Adapter
            # object and its Partition Link object always exist either both
            # or none, and that each HiperSocket Adapter can have only one
            # Partition Link.
            if num_pl == 0:
                # The Partition Link is gone, so the HS Adapter is also gone.
                self.cease_existence_local()
                raise CeasedExistence(self._uri)
            if num_pl > 1:
                # The HMC ensures that this does not happen, so if it happens
                # it is an inconsistency.
                pl_names = [pl.name for pl in part_links_for_adapter]
                raise ConsistencyError(
                    message=f"Found {num_pl} partition links "
                    f"({', '.join(pl_names)}) for adapter {self.name} on CPC "
                    f"{cpc.name} - please report this as a zhmcclient issue")
            part_link = part_links_for_adapter[0]
            part_link.delete()
            # That also deletes the Hipersocket adapter
        else:
            self.manager.session.delete(self.uri, resource=self)

        # pylint: disable=protected-access
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this Adapter.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version >= 2.13.1
        * The parent CPC of the adapter must be in DPM mode.

        Authorization requirements:

        * Object-access permission to the Adapter.
        * Task permission for the "Adapter Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'Adapter object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)

    def __repr__(self):
        """
        Return a string with the state of this Adapter, for debug purposes.
        """
        ret = (
            f"{repr_obj_id(self)} (\n"
            f"  _manager={repr_obj_id(self._manager)},\n"
            f"  _uri={self._uri!r},\n"
            f"  _full_properties={self._full_properties!r},\n"
            "  _properties_timestamp="
            f"{repr_timestamp(self._properties_timestamp)},\n"
            f"  _properties={repr_dict(self._properties, indent=4)},\n"
            f"  _ports(lazy)={repr_manager(self._ports, indent=2)}\n"
            ")")
        return ret

    @logged_api_call
    def change_crypto_type(self, crypto_type, zeroize=None):
        """
        Reconfigures a cryptographic adapter to a different crypto type.
        This operation is only supported for cryptographic adapters.

        The cryptographic adapter must be varied offline before its crypto
        type can be reconfigured.

        HMC/SE version requirements:

        * SE version >= 2.13.1
        * The parent CPC of the adapter must be in DPM mode.

        Authorization requirements:

        * Object-access permission to this Adapter.
        * Task permission to the "Adapter Details" task.

        Parameters:

          crypto_type (:term:`string`):
            - ``"accelerator"``: Crypto Express5S Accelerator
            - ``"cca-coprocessor"``: Crypto Express5S CCA Coprocessor
            - ``"ep11-coprocessor"``: Crypto Express5S EP11 Coprocessor

          zeroize (bool):
            Specifies whether the cryptographic adapter will be zeroized when
            it is reconfigured to a crypto type of ``"accelerator"``.
            `None` means that the HMC-implemented default of `True` will be
            used.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'crypto-type': crypto_type}
        if zeroize is not None:
            body['zeroize'] = zeroize
        self.manager.session.post(
            self.uri + '/operations/change-crypto-type', resource=self,
            body=body)

    @logged_api_call
    def change_adapter_type(self, adapter_type):
        """
        Reconfigures an adapter from one type to another, or to ungonfigured.
        Currently, only storage adapters can be reconfigured, and their adapter
        type is the supported storage protocol (FCP vs. FICON).

        Storage adapter instances (i.e. :class:`~zhmcclient.Adapter` objects)
        represent daughter cards on a physical storage card. Current storage
        cards require both daughter cards to be configured to the same
        protocol, so changing the type of the targeted adapter will also change
        the type of the adapter instance that represents the other daughter
        card on the same physical card. Zhmcclient users that need to determine
        the related adapter instance can do so by finding the storage adapter
        with a matching first 9 characters (card ID and slot ID) of their
        `card-location` property values.

        The targeted adapter and its related adapter on the same storage card
        must not already have the desired adapter type, they must not be
        attached to any partition, and they must not have an adapter status
        of 'exceptions'.

        HMC/SE version requirements:

        * SE version >= 2.13.1
        * The parent CPC of the adapter must be in DPM mode.

        Authorization requirements:

        * Object-access permission to this Adapter.
        * Task permission to the "Configure Storage - System Programmer" task.

        Parameters:

          adapter_type (:term:`string`):
            - ``"fcp"``: FCP (Fibre Channel Protocol)
            - ``"fc"``: FICON (Fibre Connection) protocol
            - ``"not-configured"``: No adapter type configured

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'type': adapter_type}
        self.manager.session.post(
            self.uri + '/operations/change-adapter-type', resource=self,
            body=body)

    def dump(self):
        """
        Dump this Adapter resource with its properties and child resources
        (recursively) as a resource definition.

        The returned resource definition has the following format::

            {
                # Resource properties:
                "properties": {...},

                # Child resources:
                "ports": [...],
            }

        Returns:

          dict: Resource definition of this resource.
        """

        # Dump the resource properties
        resource_dict = super().dump()

        # Dump the child resources
        ports = self.ports.dump()
        if ports:
            resource_dict['ports'] = ports

        return resource_dict

    @logged_api_call
    def list_assigned_partitions(self, full_properties=False, filter_args=None):
        """
        List the partitions assigned to this adapter.

        This method is not supported for OSA adapters configured as OSM
        (because those cannot be assigned to partitions).

        HMC/SE version requirements:

        * SE version >= 2.13.1
        * The parent CPC of the adapter must be in DPM mode.

        Authorization requirements:

        * Object-access permission to this Adapter.

        Parameters:

          full_properties (bool):
            Controls whether the full set of partition properties should be
            retrieved, vs. only a short set (uri, name, status).

          filter_args (dict):
            Filter arguments that narrow the list of returned partitions to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all assigned partitions
            are returned.

        Returns:

          : A list of :class:`~zhmcclient.Partition` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_props = ['name', 'status']
        query_parms, client_filters = divide_filter_args(
            query_props, filter_args)
        query_parms_str = make_query_str(query_parms)
        uri = (f'{self.uri}/operations/get-partitions-assigned-to-adapter'
               f'{query_parms_str}')

        result = self.manager.session.get(uri, resource=self)

        partition_mgr = self.manager.parent.partitions
        resource_obj_list = []
        for props in result['partitions-assigned-to-adapter']:

            # pylint: disable=protected-access
            resource_obj = partition_mgr.resource_class(
                manager=partition_mgr,
                uri=props[partition_mgr._uri_prop],
                name=props.get(partition_mgr._name_prop, None),
                properties=props)

            if matches_filters(resource_obj, client_filters):
                resource_obj_list.append(resource_obj)
                if full_properties:
                    resource_obj.pull_full_properties()

        return resource_obj_list

    @logged_api_call
    def list_sibling_adapters(self, full_properties=False):
        """
        List the other Adapters on the same adapter card as this Adapter.

        Some adapter cards are represented as multiple Adapter objects
        (for example, 2-port FICON Express cards, or 2-port CNA cards).
        This method lists the other Adapter objects that are on the same
        adapter card as this Adapter object.

        This is useful for example to determine the affected Adapter objects
        when replacing the adapter card, or when changing the type of a FICON
        Express adepter (see :meth:`~zhmcclient.Adapter.change_adapter_type`).

        HMC/SE version requirements:

        * SE version >= 2.13.1
        * The parent CPC of the adapter must be in DPM mode.

        Authorization requirements:

        * Object-access permission to this CPC.
        * Object-access permission to any Adapter to be included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

        Returns:

          : A list of :class:`~zhmcclient.Adapter` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # This algorithm is based on the fact that physical adapter cards
        # have their PCHIDs always within a range of 4 adjacent PCHIDs that
        # start at a multiple of 4.

        self_pchid = int(self.prop('adapter-id'), 16)
        if self_pchid >= int('7c0', 16):
            # A virtual adapter with a single PCHID -> no siblings
            return []

        # A physical adapter with a total of 4 PCHIDs reserved for the slot
        pchid_base = self_pchid // 4 * 4
        sibling_pchids = list(range(pchid_base, pchid_base + 4))
        sibling_pchids.remove(self_pchid)
        sibling_adapter_ids = [f'{p:03x}' for p in sibling_pchids]
        filter_args = {'adapter-id': sibling_adapter_ids}
        sibling_adapters = self.manager.cpc.adapters.list(
            full_properties, filter_args)
        return sibling_adapters



#!/usr/bin/env python
# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that lists partition links attached to a partition.
"""

import sys
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

if len(sys.argv) != 3:
    print(f"Usage: {sys.argv[0]} CPC PARTITION")
    sys.exit(2)

cpc_name = sys.argv[1]
part_name = sys.argv[2]

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)
    console = client.consoles.console

    print(f"Finding partition {part_name} on CPC {cpc_name} ...")
    cpc = client.cpcs.find(name=cpc_name)
    part = cpc.partitions.find(name=part_name)

    print(f"Listing partition links attached to partition {part_name} ...")
    partition_links = part.list_attached_partition_links()

    print()
    print("Partition Link        Type          State       Attached partitions")
    print("---------------------------------------------------------------------------------------")
    for pl in partition_links:
        name = pl.get_property('name')
        type = pl.get_property('type')
        state = pl.get_property('state')
        attached_parts = pl.list_attached_partitions()
        attached_part_names = [p.name for p in attached_parts]
        print(f"{name:20s}  {type:12s}  {state:10}  {', '.join(attached_part_names)}")
    print()

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for testutils._hmc_session_functions module.
"""

import pytest
from requests.packages import urllib3
import zhmcclient
import zhmcclient_mock

# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import
from zhmcclient.testutils import setup_hmc_session, teardown_hmc_session_id, \
    teardown_hmc_session, is_valid_hmc_session_id

urllib3.disable_warnings()


@pytest.mark.parametrize(
    "teardown", ['session', 'session_id']
)
def test_session_setup(hmc_definition, teardown):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test setup_hmc_session().
    """

    # The code to be tested.
    session = setup_hmc_session(hmc_definition)

    try:
        if hmc_definition.mock_file:
            assert isinstance(session, zhmcclient_mock.FakedSession)
        else:
            assert isinstance(session, zhmcclient.Session)
            assert not isinstance(session, zhmcclient_mock.FakedSession)

    finally:
        if teardown == 'session':
            teardown_hmc_session(session)
        else:
            teardown_hmc_session_id(hmc_definition, session.session_id)


@pytest.mark.parametrize(
    "teardown", ['session', 'session_id']
)
def test_session_valid_session(hmc_definition, teardown):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test is_valid_hmc_session_id() with a valid session.
    """

    session = setup_hmc_session(hmc_definition)

    try:
        # The code to be tested.
        is_valid = is_valid_hmc_session_id(hmc_definition, session.session_id)

        assert is_valid is True

    finally:
        if teardown == 'session':
            teardown_hmc_session(session)
        else:
            teardown_hmc_session_id(hmc_definition, session.session_id)


def test_session_invalid_session(hmc_definition):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test is_valid_hmc_session_id() with an invalid session.
    """
    invalid_session_id = 'invalid-session-id'

    is_valid = is_valid_hmc_session_id(hmc_definition, invalid_session_id)

    assert is_valid is False



# Copyright 2019,2021,2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper functions for HMC session management.
"""


import os
import re
import time
import logging
import pytest
import zhmcclient
import zhmcclient_mock

__all__ = ['setup_hmc_session', 'teardown_hmc_session',
           'teardown_hmc_session_id', 'is_valid_hmc_session_id',
           'LOG_FORMAT_STRING', 'LOG_DATETIME_FORMAT', 'LOG_DATETIME_TIMEZONE']

LOG_FORMAT_STRING = '%(asctime)s %(levelname)s %(name)s: %(message)s'

LOG_DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S %Z'

LOG_DATETIME_TIMEZONE = time.gmtime


def setup_hmc_session(hd):
    """
    Setup an HMC session and return a new session object for it.

    If the HMC definition represents a real HMC, log on to an HMC and return
    a new :class:`zhmcclient.Session` object.

    If the HMC definition represents a mocked HMC, create a new mock environment
    from that and return a :class:`zhmcclient_mock.FakedSession` object.

    Parameters:

      hd (:class:`~zhmcclient.testutils.HMCDefinition`): The HMC definition of
        the HMC the test runs against.

    Returns:

      :class:`~zhmcclient.Session`): Sesson object for the logged-on real or
      mocked session with the HMC.
    """
    # We use the cached skip reason from previous attempts
    skip_msg = getattr(hd, 'skip_msg', None)
    if skip_msg:
        pytest.skip(f"Skip reason from earlier attempt: {skip_msg}")

    if hd.mock_file:
        # A mocked HMC

        # Create a mocked session using the mock file from the inventory file
        session = zhmcclient_mock.FakedSession.from_hmc_yaml_file(
            hd.mock_file, userid=hd.userid, password=hd.password)

        # Set the HMC definition host to the host found in the mock file.
        hd.host = session.host

    else:
        # A real HMC

        # Enable debug logging if specified
        log_file = os.getenv('TESTLOGFILE', None)
        if log_file:

            logging.Formatter.converter = LOG_DATETIME_TIMEZONE
            log_formatter = logging.Formatter(
                LOG_FORMAT_STRING, datefmt=LOG_DATETIME_FORMAT)
            log_handler = logging.FileHandler(log_file, encoding='utf-8')
            log_handler.setFormatter(log_formatter)

            logger = logging.getLogger('zhmcclient.hmc')
            if log_handler not in logger.handlers:
                logger.addHandler(log_handler)
            logger.setLevel(logging.DEBUG)

            logger = logging.getLogger('zhmcclient.api')
            if log_handler not in logger.handlers:
                logger.addHandler(log_handler)
            logger.setLevel(logging.DEBUG)

            logger = logging.getLogger('zhmcclient.jms')
            if log_handler not in logger.handlers:
                logger.addHandler(log_handler)
            logger.setLevel(logging.DEBUG)

            logger = logging.getLogger('zhmcclient.os')
            if log_handler not in logger.handlers:
                logger.addHandler(log_handler)
            logger.setLevel(logging.DEBUG)

        rt_config = zhmcclient.RetryTimeoutConfig(
            read_timeout=1800,
        )

        # Creating a session does not interact with the HMC (logon is deferred)
        session = zhmcclient.Session(
            hd.host, hd.userid, hd.password, verify_cert=hd.verify_cert,
            retry_timeout_config=rt_config)

        # Check access to the HMC
        try:
            session.logon()
        except zhmcclient.Error as exc:
            msg = (
                f"Cannot log on to HMC {hd.nickname} at {hd.host} "
                f"due to {exc.__class__.__name__}: {exc}")
            hd.skip_msg = msg
            pytest.skip(msg)

    hd.skip_msg = None
    session.hmc_definition = hd

    return session


def teardown_hmc_session(session):
    """
    Log off from a valid session with a real HMC, identified by a zhmcclient
    session object.

    If the session represents a mocked HMC, nothing is done.

    Parameters:

      session (:class:`~zhmcclient.Session`): Session with a real or mocked HMC.
    """
    if not isinstance(session, zhmcclient_mock.FakedSession):
        session.logoff()


def teardown_hmc_session_id(hd, session_id):
    """
    Log off from a valid session with a real HMC, identified by its session ID.

    If the HMC definition represents a mocked HMC, nothing is done.

    Raises zhmcclient exceptions if the session ID is not valid.

    Parameters:

      hd (:class:`~zhmcclient.testutils.HMCDefinition`): The HMC definition of
        the HMC the test runs against.

      session_id (str): HMC session ID.
    """
    if hd.mock_file is None:
        session = zhmcclient.Session(
            hd.host, hd.userid, session_id=session_id,
            verify_cert=hd.verify_cert)
        session.logoff()


def is_valid_hmc_session_id(hd, session_id):
    """
    Return a boolean indicating whether an HMC session ID is valid.

    If the HMC definition represents a real HMC, the specified session ID
    is tested with an HMC operation.

    If the HMC definition represents a mocked HMC, no such test is performed,
    and the specified session ID is considered valid if it is `None`.

    Raises zhmcclient exceptions if the validity cannot be determined.

    Parameters:

      hd (:class:`~zhmcclient.testutils.HMCDefinition`): The HMC definition of
        the HMC the test runs against.

      session_id (str): HMC session ID to be tested for validity.
    """
    if hd.mock_file is not None:
        return session_id is None

    session = zhmcclient.Session(
        hd.host, hd.userid, session_id=session_id, verify_cert=hd.verify_cert)
    try:
        # This simply performs the GET with the session header set to the
        # session_id.
        session.get('/api/cpcs', logon_required=False, renew_session=False)
    except zhmcclient.ServerAuthError as exc:
        if re.search(r'x-api-session header did not map to a known session',
                     str(exc)):
            return False
        raise
    return True



# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The hardware messages that belong to the parent resource object (Console or
CPC).
"""


from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_HW_MESSAGE, make_query_str, timestamp_from_datetime


__all__ = ['HwMessageManager', 'HwMessage']


class HwMessageManager(BaseManager):
    """
    Manager providing access to the :term:`Hardware Messages <Hardware Message>`
    of the parent object (Console or CPC), that are exposed by the HMC this
    client is connected to.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    HMC/SE version requirements: None
    """

    def __init__(self, parent):
        # This function should not go into the docs.
        # Parameters:
        #   parent (:class:`~zhmcclient.BaseResource`):
        #      The parent object (Console or CPC).

        # Resource properties that are supported as filter query parameters
        # (for server-side filtering).
        query_props = []

        super().__init__(
            resource_class=HwMessage,
            class_name=RC_HW_MESSAGE,
            session=parent.manager.session,
            parent=parent,
            base_uri=f'{parent.uri}/hardware-messages',
            # Console: /api/console/hardware-messages/{hardware-message-id}
            # CPC: /api/cpcs/{cpc-id}/hardware-messages/{hardware-message-id}
            oid_prop='element-id',
            uri_prop='element-uri',
            name_prop='element-id',
            query_props=query_props)

    @logged_api_call
    def list(
            self, full_properties=False, filter_args=None, begin_time=None,
            end_time=None):
        """
        List the hardware messages for the parent object (Console or CPC).

        HMC/SE version requirements: None

        Authorization requirements:

        * For hardware messages of the CPC: Object-access permission to the CPC.
        * Task permission to the "Hardware Messages" task at least in view-only
          mode.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned messages to those
            whose properties match the specified filter arguments. For details,
            see :ref:`Filtering`.

            `None` causes no such filtering to happen.

          begin_time (:class:`~py:datetime.datetime`):
            Filter that narrows the list of returned messages to those created
            on or after the specified point in time.
            The datetime object may be timezone-aware or timezone-naive. If
            timezone-naive, the UTC timezone is assumed.

            `None` causes no such filtering to happen.

          end_time (:class:`~py:datetime.datetime`):
            Filter that narrows the list of returned messages to those created
            on or before the specified point in time.
            The datetime object may be timezone-aware or timezone-naive. If
            timezone-naive, the UTC timezone is assumed.

            `None` causes no such filtering to happen.

        Returns:

          list: A list of :class:`~zhmcclient.HwMessage` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.FilterConversionError`
        """
        result_prop = 'hardware-messages'
        if filter_args and 'element-id' in filter_args:
            # Filter with filter-uri instead, because faster
            element_id = filter_args['element-id']
            element_uri = f'{self._base_uri}/{element_id}'
            filter_args = dict(filter_args)
            del filter_args['element-id']
            filter_args['element-uri'] = element_uri
        query_parms = []
        if begin_time is not None:
            begin_time = timestamp_from_datetime(begin_time)
            query_parms.append(f'begin-time={begin_time}')
        if end_time is not None:
            end_time = timestamp_from_datetime(end_time)
            query_parms.append(f'end-time={end_time}')
        msg_list = self._list_with_operation(
            self._base_uri, result_prop, full_properties,
            filter_args=filter_args, query_parms=query_parms)
        if not full_properties:
            # Add the element-id property since that is used as the 'name'.
            for msg in msg_list:
                element_id = msg.uri.split('/')[-1]
                msg.update_properties_local({'element-id': element_id})
        return msg_list


class HwMessage(BaseResource):
    """
    Representation of a :term:`Hardware Message`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from list functions on their manager object
    (in this case, :class:`~zhmcclient.HwMessageManager`).

    Note that HwMessage objects do not have any writeable properties, so they
    do not have an ``update_properties()`` method.

    HMC/SE version requirements: None
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.HwMessageManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, HwMessageManager), (
            f"HwMessage init: Expected manager type {HwMessageManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    def dump(self):
        """
        Dump this HwMessage resource with its properties as a resource
        definition.

        The returned resource definition has the following format::

            {
                # Resource properties:
                "properties": {...},
            }

        Returns:

          dict: Resource definition of this resource.
        """

        # Dump the resource properties
        resource_dict = super().dump()

        return resource_dict

    # Note: HwMessage resources cannot have their properties updated,
    #       hence there is no update_properties() method.

    @logged_api_call
    def delete(self):
        """
        Delete this hardware message.

        HMC/SE version requirements: None

        Authorization requirements:

        * For hardware messages of the CPC: Object-access permission to the CPC.
        * Task permission to the "Hardware Messages" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def request_service(self, customer_name=None, customer_phone=None):
        """
        Request service from IBM for this hardware message and delete the
        hardware message.

        The hardware message's ``service-supported`` property must be True.

        HMC/SE version requirements: None

        Authorization requirements:

        * For hardware messages of the CPC: Object-access permission to the CPC.
        * Task permission to the "Hardware Messages" task.

        Parameters:

          customer_name (:term:`string`): Name of the person that can be
            contacted about the problem. Optional, default is the customer name
            registered with IBM for the machine.

          customer_phone (:term:`string`): Telephone number of the person that
            can be contacted about the problem. Optional, default is the
            customer telephone number registered with IBM for the machine.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {}
        if customer_name:
            body['customer-name'] = customer_name
        if customer_phone:
            body['customer-phone'] = customer_phone
        ops_uri = self.uri + '/operations/request-service'
        self.manager.session.post(ops_uri, resource=self, body=body)

    @logged_api_call
    def get_service_information(self, delete=False):
        """
        Get problem information and a telephone number for requesting service
        from IBM for this hardware message and optionally delete the hardware
        message.

        The hardware message's ``service-supported`` property must be True.

        HMC/SE version requirements: None

        Authorization requirements:

        * For hardware messages of the CPC: Object-access permission to the CPC.
        * Task permission to the "Hardware Messages" task.

        Parameters:

          delete (bool): Boolean indicating whether the hardware message should
            be deleted upon successful completion of the operation.

        Returns:

          dict: Response body of the "Get Console Service Request Information"
          or "Get CPC Service Request Information" operation, as described in
          the :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        query_parms = [f'delete={delete}']
        query_str = make_query_str(query_parms)
        ops_uri = f'{self.uri}/operations/get-service-information{query_str}'
        result = self.manager.session.get(ops_uri, resource=self)
        return result

    @logged_api_call
    def decline_service(self):
        """
        Decline service from IBM for this hardware message and delete the
        hardware message.

        The hardware message's ``service-supported`` property must be True.

        HMC/SE version requirements: None

        Authorization requirements:

        * For hardware messages of the CPC: Object-access permission to the CPC.
        * Task permission to the "Hardware Messages" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        ops_uri = self.uri + '/operations/request-service'
        self.manager.session.post(ops_uri, resource=self)



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for adapters (on CPCs in both DPM and classic mode).

These tests do not change any existing adapters, but create, modify
and delete Hipersocket adapters.
"""


import uuid
import warnings
import pdb
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs, all_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    standard_partition_props, runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic Adapter objects (e.g. find_by_name())
ADAPTER_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in Adapter objects returned by list() without full props
ADAPTER_LIST_PROPS = ['object-uri', 'name', 'adapter-id', 'adapter-family',
                      'type', 'status']

# Properties in Adapter objects for list(additional_properties)
ADAPTER_ADDITIONAL_PROPS = ['description', 'detected-card-type']

# Properties whose values can change between retrievals of Adapter objects
ADAPTER_VOLATILE_PROPS = []


def se_version_info(cpc):
    """
    Return the SE version of the CPC as a list of int.
    """
    return list(map(int, cpc.prop('se-version').split('.')))


def test_adapter_find_list(all_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not all_cpcs:
        pytest.skip("HMC definition does not include any CPCs")

    for cpc in all_cpcs:
        if not cpc.dpm_enabled and se_version_info(cpc) < [2, 16]:
            pytest.skip(f"CPC with SE {cpc.prop('se-version')} in classic "
                        "mode does not support Adapter objects")

        session = cpc.manager.session
        hd = session.hmc_definition

        cpc_mode_str = "DPM" if cpc.dpm_enabled else "classic"

        # Pick the adapters to test with
        adapter_list = cpc.adapters.list()
        if not adapter_list:
            skip_warn(f"No adapters on CPC {cpc.name} managed by HMC {hd.host}")
        adapter_list = pick_test_resources(adapter_list)

        for adapter in adapter_list:
            print(f"Testing on CPC {cpc.name} ({cpc_mode_str} mode) "
                  f"with adapter {adapter.name!r}")
            runtest_find_list(
                session, cpc.adapters, adapter.name, 'name', 'object-uri',
                ADAPTER_VOLATILE_PROPS, ADAPTER_MINIMAL_PROPS,
                ADAPTER_LIST_PROPS, ADAPTER_ADDITIONAL_PROPS)


def test_adapter_property(all_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not all_cpcs:
        pytest.skip("HMC definition does not include any CPCs")

    for cpc in all_cpcs:
        if not cpc.dpm_enabled and se_version_info(cpc) < [2, 16]:
            pytest.skip(f"CPC with SE {cpc.prop('se-version')} in classic "
                        "mode does not support Adapter objects")
        cpc_mode_str = "DPM" if cpc.dpm_enabled else "classic"

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the adapters to test with
        adapter_list = cpc.adapters.list()
        if not adapter_list:
            skip_warn(f"No adapters on CPC {cpc.name} ({cpc_mode_str} mode) "
                      f"managed by HMC {hd.host}")
        adapter_list = pick_test_resources(adapter_list)

        for adapter in adapter_list:
            print(f"Testing on CPC {cpc.name} ({cpc_mode_str} mode) "
                  f"with adapter {adapter.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(adapter.manager, non_list_prop)


def test_adapter_hs_crud(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a Hipersocket adapter.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        cpc_mode_str = "DPM"

        if se_version_info(cpc) >= [2, 17]:
            # TODO: Enable this case again once create_hipersocket() has been
            #       reimplemented using partition links.
            pytest.skip("create_hipersocket() is not supported on z17 CPCs")

        print(f"Testing on CPC {cpc.name} ({cpc_mode_str} mode)")

        adapter_name = TEST_PREFIX + ' test_adapter_crud adapter1'
        adapter_name_new = adapter_name + ' new'

        # Ensure a clean starting point for this test
        try:
            adapter = cpc.adapters.find(name=adapter_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                "Deleting test Hipersocket adapter from previous run: "
                f"{adapter_name!r} on CPC {cpc.name}", UserWarning)
            adapter.delete()
        try:
            adapter = cpc.adapters.find(name=adapter_name_new)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                "Deleting test Hipersocket adapter from previous run: "
                f"{adapter_name_new!r} on CPC {cpc.name}", UserWarning)
            adapter.delete()

        # Create a Hipersocket adapter
        adapter_input_props = {
            'name': adapter_name,
            'description': 'Test adapter for zhmcclient end2end tests',
        }
        adapter_auto_props = {
            'type': 'hipersockets',
        }

        # The code to be tested
        adapter = cpc.adapters.create_hipersocket(adapter_input_props)

        try:
            for pn, exp_value in adapter_input_props.items():
                assert adapter.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"
            adapter.pull_full_properties()
            for pn, exp_value in adapter_input_props.items():
                assert adapter.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"
            for pn, exp_value in adapter_auto_props.items():
                assert adapter.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"

            # Test updating a property of the adapter

            new_desc = "Updated adapter description."

            # The code to be tested
            adapter.update_properties(dict(description=new_desc))

            assert adapter.properties['description'] == new_desc
            adapter.pull_full_properties()
            assert adapter.properties['description'] == new_desc

            # Test renaming the adapter

            # The code to be tested
            adapter.update_properties(dict(name=adapter_name_new))

            assert adapter.properties['name'] == adapter_name_new
            adapter.pull_full_properties()
            assert adapter.properties['name'] == adapter_name_new
            with pytest.raises(zhmcclient.NotFound):
                cpc.adapters.find(name=adapter_name)

        finally:
            # Test deleting the adapter

            # The code to be tested
            adapter.delete()

            with pytest.raises(zhmcclient.NotFound):
                cpc.adapters.find(name=adapter_name_new)


ADAPTER_FAMILIES = [
    'hipersockets',
    'osa',
    'roce',
    'cna',
    'ficon',
    'accelerator',
    'crypto',
    'nvme',
    'coupling',
    'ism',
    'zhyperlink',
    'network-express',
    'networking',
]


@pytest.mark.parametrize(
    "test_family",
    ADAPTER_FAMILIES)
def test_adapter_list_assigned_part(dpm_mode_cpcs, test_family):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test Adapter.list_assigned_partitions().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        cpc_mode_str = "DPM"

        nes_feature = cpc.api_feature_enabled('network-express-support')

        # Determine the list of adapters on the CPC with the test family
        family_adapters = []
        adapter_list = cpc.adapters.list()
        for adapter in adapter_list:
            family = adapter.get_property('adapter-family')
            if family not in ADAPTER_FAMILIES:
                warnings.warn(
                    f"Ignoring adapter {adapter.name!r} on CPC {cpc.name} with "
                    f"an unknown family: '{family}'", UserWarning)
                continue
            if family != test_family:
                continue
            family_adapters.append(adapter)

        if not family_adapters:
            pytest.skip(f"CPC {cpc.name} does not have any adapters with "
                        f"family: {test_family} ")

        tmp_part = None
        try:

            if test_family in ('hipersockets', 'osa') and not nes_feature:
                # Vswitch-based adapter

                # Skip OSM adapters since they cannot be assigned to partitions
                # and the "Get Partitions Assigned to Adapter" operation is not
                # supported for OSM adapters.
                eligible_adapters = []
                for adapter in family_adapters:
                    if adapter.get_property('type') == 'osm':
                        continue
                    eligible_adapters.append(adapter)

                # Create a temporary partition for test purposes
                part_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
                part_props = standard_partition_props(cpc, part_name)
                tmp_part = cpc.partitions.create(part_props)

                test_adapters = pick_test_resources(eligible_adapters)
                for test_adapter in test_adapters:

                    print(f"Testing on CPC {cpc.name} ({cpc_mode_str} mode) "
                          "with vswitch-based network adapter "
                          f"{test_adapter.name!r} (family '{test_family}')")

                    # The method to be tested
                    before_parts = test_adapter.list_assigned_partitions()

                    # Find the virtual switch for the test adapter
                    filter_args = {'backing-adapter-uri': test_adapter.uri}
                    vswitches = cpc.virtual_switches.list(
                        filter_args=filter_args)
                    assert len(vswitches) >= 1
                    vswitch = vswitches[0]

                    # Create a NIC in the temporary partition
                    nic_name = f"{family}_{uuid.uuid4().hex}"
                    nic_props = {
                        'name': nic_name,
                        'virtual-switch-uri': vswitch.uri,
                    }
                    tmp_part.nics.create(nic_props)

                    # The method to be tested
                    after_parts = test_adapter.list_assigned_partitions()

                    before_uris = [p.uri for p in before_parts]
                    new_parts = []
                    for part in after_parts:
                        if part.uri not in before_uris:
                            new_parts.append(part)

                    assert len(new_parts) == 1
                    new_part = new_parts[0]
                    assert new_part.uri == tmp_part.uri

            elif test_family in ('hipersockets', 'osa') and nes_feature or \
                    test_family in ('roce', 'cna', 'network-express',
                                    'networking'):
                # Port-based adapter

                # Create a temporary partition for test purposes
                part_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
                part_props = standard_partition_props(cpc, part_name)
                tmp_part = cpc.partitions.create(part_props)

                test_adapters = pick_test_resources(family_adapters)
                for test_adapter in test_adapters:

                    print(f"Testing on CPC {cpc.name} ({cpc_mode_str} mode) "
                          "with port-based network adapter "
                          f"{test_adapter.name!r} (family '{test_family}')")

                    # The method to be tested
                    before_parts = test_adapter.list_assigned_partitions()

                    test_port = test_adapter.ports.list()[0]

                    # Create a NIC in the temporary partition
                    nic_name = f"{family}_{uuid.uuid4().hex}"
                    nic_props = {
                        'name': nic_name,
                        'network-adapter-port-uri': test_port.uri,
                    }
                    if test_family == 'network-express':
                        nic_props['type'] = "neth"
                    tmp_part.nics.create(nic_props)

                    # The method to be tested
                    after_parts = test_adapter.list_assigned_partitions()

                    before_uris = [p.uri for p in before_parts]
                    new_parts = []
                    for part in after_parts:
                        if part.uri not in before_uris:
                            new_parts.append(part)

                    assert len(new_parts) == 1
                    new_part = new_parts[0]
                    assert new_part.uri == tmp_part.uri

            elif test_family == 'ficon':

                # This test works only for FCP adapters.
                fcp_adapters = []
                for adapter in family_adapters:
                    if adapter.get_property('type') != 'fcp':
                        continue
                    fcp_adapters.append(adapter)

                # Assigning storage adapters to partitions is complex,
                # so we search for a storage adapter that is already
                # assigned to a partition.
                assigned_fcp_adapters = []
                for adapter in fcp_adapters:

                    # The method to be tested
                    assigned_parts = adapter.list_assigned_partitions()

                    if len(assigned_parts) == 0:
                        continue
                    assigned_fcp_adapters.append(adapter)

                if not assigned_fcp_adapters:
                    pytest.skip(f"CPC {cpc.name} does not have FCP adapters "
                                "assigned to partitions")

                test_adapters = pick_test_resources(assigned_fcp_adapters)
                for test_adapter in test_adapters:

                    print(f"Testing on CPC {cpc.name} ({cpc_mode_str} mode) "
                          f"with FCP adapter "
                          f"{test_adapter.name!r} (family '{test_family}')")

                    # The method to be tested
                    before_parts = test_adapter.list_assigned_partitions()

                    found_adapters = []
                    for part in before_parts:
                        sgroups = part.list_attached_storage_groups()
                        for sg in sgroups:
                            if sg.get_property('type') != 'fcp':
                                # This test works only for FCP stogrps
                                continue
                            vsrs = sg.virtual_storage_resources.list()
                            for vsr in vsrs:
                                port = vsr.adapter_port
                                adapter = port.manager.parent
                                if adapter.uri == test_adapter.uri:
                                    found_adapters.append(adapter)

                    assert len(found_adapters) >= 1

            elif test_family == 'accelerator':

                # Create a temporary partition for test purposes
                part_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
                part_props = standard_partition_props(cpc, part_name)
                tmp_part = cpc.partitions.create(part_props)

                test_adapters = pick_test_resources(family_adapters)
                for test_adapter in test_adapters:

                    print(f"Testing on CPC {cpc.name} ({cpc_mode_str} mode) "
                          "with accelerator adapter "
                          f"{test_adapter.name!r} (family '{test_family}')")

                    # The method to be tested
                    before_parts = test_adapter.list_assigned_partitions()

                    # Create a VF in the temporary partition
                    vf_name = f"{family}_{uuid.uuid4().hex}"
                    vf_props = {
                        'name': vf_name,
                        'adapter-uri': test_adapter.uri,
                    }
                    tmp_part.virtual_functions.create(vf_props)

                    # The method to be tested
                    after_parts = test_adapter.list_assigned_partitions()

                    before_uris = [p.uri for p in before_parts]
                    new_parts = []
                    for part in after_parts:
                        if part.uri not in before_uris:
                            new_parts.append(part)

                    assert len(new_parts) == 1
                    new_part = new_parts[0]
                    assert new_part.uri == tmp_part.uri

            elif test_family == 'crypto':

                # Assigning crypto adapters to partitions is complex,
                # so we search for a crypto adapter that is already
                # assigned to a partition.
                assigned_crypto_adapters = []
                for adapter in family_adapters:

                    # The method to be tested
                    assigned_parts = adapter.list_assigned_partitions()

                    if len(assigned_parts) == 0:
                        continue
                    assigned_crypto_adapters.append(adapter)

                if not assigned_crypto_adapters:
                    pytest.skip(f"CPC {cpc.name} does not have crypto adapters "
                                "assigned to partitions")

                test_adapters = pick_test_resources(assigned_crypto_adapters)
                for test_adapter in test_adapters:

                    print(f"Testing on CPC {cpc.name} ({cpc_mode_str} mode) "
                          "with crypto adapter "
                          f"{test_adapter.name!r} (family '{test_family}')")

                    # The method to be tested
                    before_parts = test_adapter.list_assigned_partitions()

                    found_adapter_uris = []
                    for part in before_parts:
                        crypto_config = part.get_property(
                            'crypto-configuration')
                        if crypto_config is None:
                            continue
                        adapter_uris = crypto_config['crypto-adapter-uris']
                        for adapter_uri in adapter_uris:
                            if adapter_uri == test_adapter.uri:
                                found_adapter_uris.append(adapter_uri)

                    assert len(found_adapter_uris) >= 1

            elif test_family == 'nvme':
                print(f"TODO: Implement test for family {test_family}")

            elif test_family == 'coupling':
                print(f"TODO: Implement test for family {test_family}")

            elif test_family == 'ism':
                print(f"TODO: Implement test for family {test_family}")

            elif test_family == 'zhyperlink':
                print(f"TODO: Implement test for family {test_family}")

            else:
                assert False, f"Unexpected family {test_family}"

        finally:
            # Cleanup
            if tmp_part:
                tmp_part.delete()


def base_adapter_id(adapter_id, family):
    """
    Return the base adapter_id of an adapter_id.
    Both are in hex.
    """

    # Adapter families that have no physical adapter card
    virtual_families = [
        'hipersockets',
        'coupling',  # Internal Coupling Facility
        'ism',  # Internal Shared Memory
    ]

    pchid = int(adapter_id, 16)

    if family in virtual_families:

        # Make sure the virtual card detection based on the PCHID range in
        # Adapter.list_sibling_adapters() holds true
        assert pchid >= int('7c0', 16)

        # Virtual adapters have only 1 PCHID
        base_pchid = pchid

    else:

        # Make sure the physical card detection based on the PCHID range in
        # Adapter.list_sibling_adapters() holds true
        assert pchid < int('7c0', 16)

        # Physical adapters have 4 PCHIDs reeserved for the slot.
        # Calculate the base PCHID for the slot.
        base_pchid = pchid // 4 * 4

    return f'{base_pchid:03x}'


def test_adapter_list_sibling_adapters(all_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test Adapter.list_sibling_adapters().
    """
    if not all_cpcs:
        pytest.skip("HMC definition does not include any CPCs")

    for cpc in all_cpcs:
        if not cpc.dpm_enabled and se_version_info(cpc) < [2, 16]:
            pytest.skip(f"CPC with SE {cpc.prop('se-version')} in classic "
                        "mode does not support Adapter objects")

        adapters = cpc.adapters.list()

        # Group the adapters by their base PCHID
        adapters_by_base = {}  # key: base PCHID, value: list of Adapter
        for adapter in adapters:
            adapter_id = adapter.get_property('adapter-id')
            family = adapter.get_property('adapter-family')
            base_id = base_adapter_id(adapter_id, family)
            if base_id not in adapters_by_base:
                adapters_by_base[base_id] = []
            adapters_by_base[base_id].append(adapter)

        # Test all adapters for their siblings
        for adapter in adapters:
            adapter_id = adapter.get_property('adapter-id')
            family = adapter.get_property('adapter-family')
            base_id = base_adapter_id(adapter_id, family)

            # The code to be tested
            sibling_adapters = adapter.list_sibling_adapters()

            assert isinstance(sibling_adapters, list)

            sibling_ids = {a.get_property('adapter-id')
                           for a in sibling_adapters}
            sibling_names = [a.name for a in sibling_adapters]
            exp_ids = {a.get_property('adapter-id')
                       for a in adapters_by_base[base_id]}
            exp_ids.remove(adapter_id)
            exp_names = [a.name for a in adapters_by_base[base_id]]
            exp_names.remove(adapter.name)

            assert sibling_ids == exp_ids, (
                f"Adapter '{adapter.name}' has unexpected siblings: got: "
                f"{sibling_names}, expected: {exp_names}")


# Properties in returned adapters, where each list item is a
# tuple(prop name, minimum CPC version).

# Default properties when no full/additional properties are requested:
DEFAULT_PROPS_LIST_PERMITTED_ADAPTERS = [
    ('object-uri', [2, 13, 1]),
    ('name', [2, 13, 1]),
    ('adapter-id', [2, 13, 1]),
    ('adapter-family', [2, 13, 1]),
    ('type', [2, 13, 1]),
    ('status', [2, 13, 1]),
    # The 'firmware-update-pending' property requires the
    # LI_1580_CRYPTO_AUTO_TOGGLE feature to be enabled:
    ('firmware-update-pending', [2, 16, 0]),
    ('cpc-name', [2, 13, 1]),
    ('cpc-object-uri', [2, 13, 1]),
    ('se-version', [2, 13, 1]),
    ('dpm-enabled', [2, 13, 1]),
]

# Full set of properties that are common on all types of adapters:
COMMON_FULL_PROPS_ADAPTERS = DEFAULT_PROPS_LIST_PERMITTED_ADAPTERS + [
    ('object-id', [2, 13, 1]),
    ('class', [2, 13, 1]),
    ('parent', [2, 13, 1]),
    ('description', [2, 13, 1]),
    ('detected-card-type', [2, 13, 1]),
    ('state', [2, 13, 1]),
    ('physical-channel-status', [2, 13, 1]),
]

LIST_PERMITTED_ADAPTERS_TESTCASES = [
    # The list items are tuples with the following items:
    # - desc (string): description of the testcase.
    # - input_kwargs (dict): Input parameters for the function.
    # - exp_props (list): List of expected properties in the result. Each list
    #     item is a tuple(prop name, minimum CPC version).
    # - exp_exc_type (class): Expected exception type, or None for success.
    # - run (bool or 'pdb'): Whether to run the test or call the debugger.

    (
        "Default parameters",
        dict(),
        DEFAULT_PROPS_LIST_PERMITTED_ADAPTERS,
        None,
        True,
    ),
    (
        "full_properties",
        dict(
            full_properties=True,
        ),
        COMMON_FULL_PROPS_ADAPTERS,
        None,
        True,
    ),
    (
        "'detected-card-type' in additional_properties",
        dict(
            additional_properties=['detected-card-type'],
        ),
        DEFAULT_PROPS_LIST_PERMITTED_ADAPTERS + [
            ('detected-card-type', [2, 13, 1]),
        ],
        None,
        True,
    ),
]


@pytest.mark.parametrize(
    "desc, input_kwargs, exp_props, exp_exc_type, run",
    LIST_PERMITTED_ADAPTERS_TESTCASES)
def test_adapter_list_permitted(
        desc, input_kwargs, exp_props, exp_exc_type, run,
        all_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name, unused-argument
    """
    Test Console.list_permitted_adapters() without filtering, but with
    different variations of returned properties.
    """
    if not all_cpcs:
        pytest.skip("HMC definition does not include any CPCs")

    console = all_cpcs[0].manager.console
    session = console.manager.session
    client = console.manager.client
    hd = session.hmc_definition

    if hd.mock_file:
        skip_warn("zhmcclient mock does not support 'List Permitted Adapters' "
                  "operation")

    if run == 'pdb':
        # pylint: disable=forgotten-debug-statement
        pdb.set_trace()

    if not run:
        skip_warn("Testcase is disabled in testcase definition")

    # Prepare what the managed CPCs support and from which CPCs adapters
    # are expected.
    supports_full_properties = True
    supports_add_properties = True
    expected_cpc_names = set()
    cpc_by_name = {}
    cpc_list = client.cpcs.list()
    for cpc in cpc_list:
        cpc_version_str = cpc.properties['se-version']
        cpc_version = list(map(int, cpc_version_str.split('.')))
        if not cpc.dpm_enabled:
            supports_full_properties = False
        if cpc.dpm_enabled or cpc_version >= [2, 16]:
            expected_cpc_names.add(cpc.name)
        if not cpc.api_feature_enabled('adapter-network-info'):
            supports_add_properties = False
        cpc_by_name[cpc.name] = cpc

    if 'full_properties' in input_kwargs and not supports_full_properties:
        skip_warn("The managed CPCs do not support full_properties=True in "
                  "list_permitted_adapters()")
    if 'additional_properties' in input_kwargs and not supports_add_properties:
        skip_warn("The managed CPCs do not support additional_properties=True "
                  "in list_permitted_adapters()")

    if exp_exc_type:

        with pytest.raises(exp_exc_type) as exc_info:

            # Exercise the code to be tested
            _ = console.list_permitted_adapters(**input_kwargs)

        _ = exc_info.value
    else:

        # Exercise the code to be tested
        adapters = console.list_permitted_adapters(**input_kwargs)

        # Verify the result
        returned_cpc_names = set()
        for adapter in adapters:
            cpc = adapter.manager.parent
            returned_cpc_names.add(cpc.name)
            cpc_with_props = cpc_by_name[cpc.name]
            cpc_version_str = cpc_with_props.properties['se-version']
            cpc_version = list(map(int, cpc_version_str.split('.')))
            assert isinstance(adapter, zhmcclient.Adapter)
            for pname, min_cpc_version in exp_props:
                if cpc_version >= min_cpc_version:
                    # The property is supposed to be in the result
                    actual_pnames = list(adapter.properties.keys())
                    if pname == 'firmware-update-pending' \
                            and pname not in actual_pnames:
                        warnings.warn(
                            "The 'firmware-update-pending' property is not "
                            f"returned for adapters on CPC {cpc.name!r} (SE "
                            f"version {cpc_version}). Check on the "
                            "SE 'Manage Firmware Features' task whether "
                            "feature LI_1580_CRYPTO_AUTO_TOGGLE is enabled.")
                    else:
                        assert pname in actual_pnames, (
                            f"Expected property {pname!r} is not in the "
                            f"properties of adapter: {adapter!r}")

        # Verify that adapters are returned:
        assert returned_cpc_names == expected_cpc_names, (
            "Invalid set of CPCs of returned adapters:\n"
            f"Returned CPCs: {','.join(returned_cpc_names)}"
            f"Expected CPCs: {','.join(expected_cpc_names)}")



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for partitions (on CPCs in DPM mode).

These tests do not change any existing partitions, but create, modify and delete
test partitions.
"""


import warnings
import random
import uuid
from datetime import timedelta, datetime, timezone
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
from .utils import logger  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    standard_partition_props, runtest_find_list, runtest_get_properties, \
    pformat_as_dict, assert_res_prop, validate_firmware_features

urllib3.disable_warnings()

# Properties in minimalistic Partition objects (e.g. find_by_name())
PART_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in Partition objects returned by list() without full props
PART_LIST_PROPS = ['object-uri', 'name', 'status', 'type']

# Properties in Partition objects for list(additional_properties)
PART_ADDITIONAL_PROPS = ['description', 'short-name']

# Properties whose values can change between retrievals of Partition objects
PART_VOLATILE_PROPS = []


def assert_ctc_partition_link(partlink, num_paths, num_partitions):
    """
    Assert that the partition link currently has the expected number of
    connections between partitions, and the expected number of paths.
    Each partition must be connected to every other partition.
    """
    paths = partlink.get_property('paths')

    act_num_paths = len(paths)
    assert act_num_paths == num_paths, (
        "Unexpected number of paths in CTC-type partition link "
        f"{partlink.name!r}:\n"
        f"  Expected: {num_paths}\n"
        f"  Actual: {act_num_paths}\n"
        "Partition link properties:\n"
        f"{pformat_as_dict(partlink.properties)}")

    # Number of connections between all partitions
    num_devices = int(num_partitions * (num_partitions - 1) / 2)

    for path in paths:
        act_num_devices = len(path['devices'])
        assert act_num_devices == num_devices, (
            "Unexpected number of partition connections in CTC-type "
            f"partition link {partlink.name!r}:\n"
            f"  Expected: {num_devices}\n"
            f"  Actual: {act_num_devices}\n"
            "Partition link properties:\n"
            f"{pformat_as_dict(partlink.properties)}")


def test_part_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the partitions to test with
        part_list = cpc.partitions.list()
        if not part_list:
            skip_warn(f"No partitions on CPC {cpc.name} managed by "
                      f"HMC {hd.host}")
        part_list = pick_test_resources(part_list)

        for part in part_list:
            print(f"Testing on CPC {cpc.name} with partition {part.name!r}")
            runtest_find_list(
                session, cpc.partitions, part.name, 'name', 'status',
                PART_VOLATILE_PROPS, PART_MINIMAL_PROPS, PART_LIST_PROPS,
                PART_ADDITIONAL_PROPS)


def test_part_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the partitions to test with
        part_list = cpc.partitions.list()
        if not part_list:
            skip_warn(f"No partitions on CPC {cpc.name} managed by HMC "
                      f"{hd.host}")
        part_list = pick_test_resources(part_list)

        for part in part_list:
            print(f"Testing on CPC {cpc.name} with partition {part.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(part.manager, non_list_prop)


def test_part_crud(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a partition.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        print(f"Testing on CPC {cpc.name}")

        part_name = TEST_PREFIX + ' test_part_crud part1'
        part_name_new = part_name + ' new'

        # Ensure a clean starting point for this test
        try:
            part = cpc.partitions.find(name=part_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                f"Deleting test partition from previous run: {part_name!r} on "
                f"CPC {cpc.name}", UserWarning)
            status = part.get_property('status')
            if status != 'stopped':
                part.stop()
            part.delete()
        try:
            part = cpc.partitions.find(name=part_name_new)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                f"Deleting test partition from previous run: {part_name_new!r} "
                f"on CPC {cpc.name}", UserWarning)
            status = part.get_property('status')
            if status != 'stopped':
                part.stop()
            part.delete()

        # Test creating the partition

        part_input_props = standard_partition_props(cpc, part_name)
        part_auto_props = {
            'status': 'stopped',
        }

        # The code to be tested
        part = cpc.partitions.create(part_input_props)

        for pn, exp_value in part_input_props.items():
            assert part.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"
        part.pull_full_properties()
        for pn, exp_value in part_input_props.items():
            assert part.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"
        for pn, exp_value in part_auto_props.items():
            assert part.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"

        # Test updating a property of the partition

        new_desc = "Updated partition description."

        # The code to be tested
        part.update_properties(dict(description=new_desc))

        assert part.properties['description'] == new_desc
        part.pull_full_properties()
        assert part.properties['description'] == new_desc

        # Test renaming the partition

        # The code to be tested
        part.update_properties(dict(name=part_name_new))

        assert part.properties['name'] == part_name_new
        part.pull_full_properties()
        assert part.properties['name'] == part_name_new
        with pytest.raises(zhmcclient.NotFound):
            cpc.partitions.find(name=part_name)

        # Test deleting the partition

        # The code to be tested
        part.delete()

        with pytest.raises(zhmcclient.NotFound):
            cpc.partitions.find(name=part_name_new)


def test_partition_features(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test features of the CPC of a partition:
    - For firmware features:
      - feature_enabled() - deprecated
      - firmware_feature_enabled()
      - feature_info()
      - list_firmware_features()
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the partition to test with
        part_list = cpc.partitions.list()
        if not part_list:
            skip_warn(
                f"No partitions on CPC {cpc.name} managed by HMC {hd.host}")
        part = random.choice(part_list)

        client = cpc.manager.client
        api_version_info = client.version_info()

        cpc.pull_properties(['available-features-list'])
        cpc_fw_features = cpc.prop('available-features-list', None)

        part.pull_properties(['available-features-list'])
        part_fw_features = part.prop('available-features-list', None)

        fw_feature_name = 'dpm-storage-management'
        if cpc_fw_features is None:
            # The machine does not yet support firmware features
            with pytest.raises(ValueError):
                # The code to be tested: feature_enabled()
                part.feature_enabled(fw_feature_name)
            enabled = part.firmware_feature_enabled(fw_feature_name)
            assert enabled is False
        else:

            # Verify that the feature lists of Partition and CPC are equal
            assert len(part_fw_features) == len(cpc_fw_features)
            part_state_by_name = {f['name']: f['state']
                                  for f in part_fw_features}
            for cpc_feature in cpc_fw_features:
                name = cpc_feature['name']
                assert name in part_state_by_name, (
                    f"Firmware feature {name!r} is available for CPC but not "
                    f"for partition {part.name!r}")
                part_enabled = part_state_by_name[name]
                cpc_enabled = cpc_feature['state']
                assert part_enabled == cpc_enabled, (
                    f"Firmware feature {name!r} has different enablement state "
                    f"on CPC {cpc_enabled} than on partition {part.name!r} "
                    f"({part_enabled})")

            if fw_feature_name not in part_state_by_name:
                # The machine generally supports firmware features, but this
                # feature is not available
                with pytest.raises(ValueError):
                    # The code to be tested: feature_enabled()
                    part.feature_enabled(fw_feature_name)
                # The code to be tested: firmware_feature_enabled()
                enabled = part.firmware_feature_enabled(fw_feature_name)
                assert enabled is False
            else:
                # The machine has this feature available
                # The code to be tested: feature_enabled()
                enabled = part.feature_enabled(fw_feature_name)
                exp_enabled = part_state_by_name[fw_feature_name]
                assert_res_prop(enabled, exp_enabled,
                                'available-features-list', cpc)
                # The code to be tested: firmware_feature_enabled()
                enabled = part.firmware_feature_enabled(fw_feature_name)
                assert enabled == exp_enabled
        # Test: feature_info()
        if cpc_fw_features is None:
            # The machine does not yet support features
            with pytest.raises(ValueError):
                # The code to be tested: feature_info()
                part.feature_info()
        else:
            # The machine supports features
            # The code to be tested: feature_info()
            features = part.feature_info()
            # Note: It is possible that the feature list exists but is empty
            #       (e.g when a z14 HMC manages a z13)
            for i, feature in enumerate(features):
                assert 'name' in feature, (
                    f"Feature #{i} does not have the 'name' attribute "
                    f"in Partition object for partition {part.name}")
                assert 'description' in feature, (
                    f"Feature #{i} does not have the 'description' attribute "
                    f"in Partition object for partition {part.name}")
                assert 'state' in feature, (
                    f"Feature #{i} does not have the 'state' attribute "
                    f"in Partition object for partition {part.name}")
        # The code to be tested: list_firmware_features()
        fw_features = part.list_firmware_features()
        validate_firmware_features(api_version_info, fw_features)


def test_part_list_os_messages(logger, dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test "List OS Messages" operation on partitions
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        if hd.mock_file:
            skip_warn("zhmcclient mock does not support 'List OS Messages' "
                      "operation")

        # Pick the partition to test with
        part_list = cpc.partitions.list()
        if not part_list:
            skip_warn(
                f"No partitions on CPC {cpc.name} managed by HMC {hd.host}")

        test_part = None
        for part in part_list:

            # Test: List all messages (without begin or end)
            try:
                logger.debug(
                    "Listing OS messages of partition %s with no begin/end",
                    part.name)
                result = part.list_os_messages()
            except zhmcclient.HTTPError as exc:
                if exc.http_status == 409 and exc.reason == 332:
                    # Meaning: The messages interface for the partition is not
                    # available
                    logger.debug(
                        "Partition %s cannot list OS messages", part.name)
                    continue
                raise

            all_messages = result['os-messages']
            if len(all_messages) >= 3:
                test_part = part
                break

            logger.debug(
                "Partition %s has only %d OS messages",
                part.name, len(all_messages))

        if test_part is None:
            skip_warn(f"No partition on CPC {cpc.name} has the minimum number "
                      "of 3 OS messages for the test")

        # Test with begin/end selecting the full set of messages
        all_begin = all_messages[0]['sequence-number']
        all_end = all_messages[-1]['sequence-number']
        logger.debug(
            "Listing OS messages of partition %s with begin=%s, end=%s",
            test_part.name, all_begin, all_end)
        result = test_part.list_os_messages(begin=all_begin, end=all_end)
        messages = result['os-messages']
        assert len(messages) == all_end - all_begin + 1
        assert messages == all_messages

        # Test with begin/end selecting a subset of messages
        while True:
            seq1 = random.choice(all_messages)['sequence-number']
            if seq1 not in {all_begin, all_end}:
                break
        while True:
            seq2 = random.choice(all_messages)['sequence-number']
            if seq2 not in {all_begin, all_end, seq1}:
                break
        begin = min(seq1, seq2)
        end = max(seq1, seq2)
        logger.debug(
            "Listing OS messages of partition %s with begin=%s, end=%s",
            test_part.name, begin, end)
        result = test_part.list_os_messages(begin=begin, end=end)
        messages = result['os-messages']
        assert len(messages) == end - begin + 1
        for message in messages:
            assert message in all_messages


def test_part_create_os_websocket(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test "Get ASCII Console WebSocket URI" operation on partitions
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        if hd.mock_file:
            skip_warn("zhmcclient mock does not support 'Get ASCII Console "
                      "WebSocket URI' operation")

        # Pick the partition to test with
        active_part_list = cpc.partitions.list(filter_args={'status': 'active'})
        if not active_part_list:
            skip_warn(
                f"No partitions on CPC {cpc.name} managed by HMC {hd.host} "
                "with an active status")

        # Pick a random partition to test with
        part = random.choice(active_part_list)

        ws_uri = part.create_os_websocket()

        assert ws_uri.startswith('/api/websock/')


# Full set of properties that are common on all types of partitions:
COMMON_PROPS_LIST = ['name', 'object-uri', 'type', 'status',
                     'has-unacceptable-status']
FULL_PROPS_LIST = COMMON_PROPS_LIST + [
    'initial-memory', 'maximum-memory', 'partition-id', 'processor-mode',
    'autogenerate-partition-id', 'boot-device', 'boot-network-device',
    'description', 'hba-uris', 'ifl-processors',
    'initial-ifl-processing-weight', 'nic-uris', 'partition-link-uris',
    'storage-group-uris', 'tape-link-uris', 'virtual-function-uris'
]
LIST_PERMITTED_PARTITION_TESTCASES = [
    # The list items are tuples with the following items:
    # - desc (string): description of the testcase.
    # - input_kwargs (dict): Input parameters for the function.
    # - exp_props (List) : Expected properties in the output
    (
        "Default parameters",
        dict(),
        COMMON_PROPS_LIST,
    ),
    (
        "full_properties",
        dict(
            full_properties=True,
        ),
        FULL_PROPS_LIST
    ),
    (
        "full_properties and filtering",
        dict(
            full_properties=True,
            filter_args={'type': 'linux'},
        ),
        FULL_PROPS_LIST
    ),
    (
        "Bad request in filter_args",
        dict(
            filter_args={'name': '@#1c'}
        ),
        COMMON_PROPS_LIST
    ),
    (
        "additional-properties",
        dict(
            additional_properties=['partition-id', 'tape-link-uris']
        ),
        COMMON_PROPS_LIST + ['partition-id', 'tape-link-uris']
    ),
]


@pytest.mark.parametrize(
    "desc, input_kwargs, exp_props",
    LIST_PERMITTED_PARTITION_TESTCASES)
def test_console_list_permitted_partitions(desc, input_kwargs, exp_props,
                                           dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name, unused-argument
    """
    Test list permitted partitions on a cpc
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        session = cpc.manager.session
        hd = session.hmc_definition
        client = zhmcclient.Client(session)
        console = client.consoles.console
        features = console.list_api_features()
        if 'additional_properties' in input_kwargs and \
                'dpm-ctc-partition-link-management' not in features and \
                'dpm-hipersockets-partition-link-management' not in features:
            pytest.skip("HMC does not support additional-properties parameter.")

        permitted_part_list = console.list_permitted_partitions(**input_kwargs)
        if not permitted_part_list:
            skip_warn(f"No partitions on CPC {cpc.name} managed by HMC "
                      f"{hd.host} for the user {session.userid}")

        permitted_part_list = pick_test_resources(permitted_part_list)
        for partition in permitted_part_list:
            assert isinstance(partition, zhmcclient.Partition)
            for pname in exp_props:
                # The property is supposed to be in the result
                actual_pnames = list(partition.properties.keys())
                assert pname in actual_pnames, \
                    f"Actual partition: {partition!r}"

        # Test list permitted partitions with filtering
        permitted_partitions_filter_agrs = (console.list_permitted_partitions(
            filter_args={'name': permitted_part_list[0].name,
                         'cpc-name': permitted_part_list[
                             0].manager.parent.name}))
        assert len(permitted_partitions_filter_agrs) == 1

        # Test list permitted partitions with bad filter args
        permitted_partitions_filter_agrs = (console.list_permitted_partitions(
            filter_args={'name': permitted_part_list[0].name,
                         'cpc-name': 'bad-cpc'}))
        assert len(permitted_partitions_filter_agrs) == 0


TESTCASES_PART_GET_SUSTAINABILITY_DATA = [
    # Test cases for test_part_get_sustainability_data(), each as a tuple with
    # these items:
    # * tc: Testcase short description
    # * input_kwargs: kwargs to be used as input parameters for
    #   Partition.get_sustainability_data()
    # * exp_oldest: expected delta time from now to oldest data point,
    #   as timedelta
    # * exp_delta: expected delta time between data points, as timedelta
    (
        "Default values (range=last-week, resolution=one-hour)",
        {},
        timedelta(days=7),
        timedelta(hours=1),
    ),
    (
        "range=last-day, resolution=one-hour",
        {
            "range": "last-day",
            "resolution": "one-hour",
        },
        timedelta(hours=24),
        timedelta(hours=1),
    ),
    (
        "range=last-day, resolution=fifteen-minutes",
        {
            "range": "last-day",
            "resolution": "fifteen-minutes",
        },
        timedelta(hours=24),
        timedelta(minutes=15),
    ),
]

PART_METRICS = {
    # Metrics returned in "Get Partition Historical Sustainability Data" HMC
    # operation.
    # metric name: data type
    "wattage": int,
    "processor-utilization": int,
}


@pytest.mark.parametrize(
    "tc, input_kwargs, exp_oldest, exp_delta",
    TESTCASES_PART_GET_SUSTAINABILITY_DATA)
def test_part_get_sustainability_data(
        tc, input_kwargs, exp_oldest, exp_delta,
        dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Test for Partition.get_sustainability_data(...)
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        if hd.mock_file:
            skip_warn("zhmcclient mock does not support "
                      "Partition.get_sustainability_data()")

        # Pick the partition to test with
        part_list = cpc.partitions.list()
        if not part_list:
            skip_warn(
                f"No partitions on CPC {cpc.name} managed by HMC {hd.host}")

        # Pick a random partition to test with
        part = random.choice(part_list)

        session = part.manager.session
        hd = session.hmc_definition

        print(f"Testing with partition {part.name}")

        try:

            # The code to be tested
            data = part.get_sustainability_data(**input_kwargs)

        except zhmcclient.HTTPError as exc:
            if exc.http_status == 403 and exc.reason == 1:
                skip_warn(
                    f"HMC userid {hd.userid!r} is not authorized for task "
                    f"'Environmental Dashboard' on HMC {hd.host}")
            elif exc.http_status == 404 and exc.reason == 1:
                skip_warn(
                    f"Partition {part.name} on HMC {hd.host} does not support "
                    f"feature: {exc}")
            else:
                raise

        now_dt = datetime.now(timezone.utc)
        exp_oldest_dt = now_dt - exp_oldest

        act_metric_names = set(data.keys())
        exp_metric_names = set(PART_METRICS.keys())
        assert act_metric_names == exp_metric_names

        for metric_name, metric_array in data.items():
            metric_type = PART_METRICS[metric_name]

            first_item = True
            previous_dt = None
            for dp_item in metric_array:
                # We assume the order is oldest to newest

                assert 'data' in dp_item
                assert 'timestamp' in dp_item
                assert len(dp_item) == 2

                dp_data = dp_item['data']
                dp_timestamp_dt = dp_item['timestamp']

                assert isinstance(dp_data, metric_type), \
                    f"Invalid data type for metric {metric_name!r}"

                if first_item:
                    first_item = False

                    # Verify that the oldest timestamp is within a certain
                    # delta from the range start.
                    # There are cases where that is not satisfied, so we only
                    # issue only a warning (as opposed to failing).
                    delta_sec = abs((dp_timestamp_dt - exp_oldest_dt).seconds)
                    if delta_sec > 15 * 60:
                        print(f"Warning: Oldest data point of metric "
                              f"{metric_name!r} is not within 15 minutes of "
                              "range start: "
                              f"Oldest data point: {dp_timestamp_dt}, "
                              f"Range start: {exp_oldest_dt}, "
                              f"Delta: {delta_sec} sec")
                else:

                    # For second oldest timestamp on, verify that the delta
                    # to the previous data point is the requested resolution.
                    # There are cases where that is not satisfied, so we only
                    # issue only a warning (as opposed to failing).
                    tolerance_pct = 10
                    delta_td = abs(dp_timestamp_dt - previous_dt)
                    if abs(delta_td.seconds - exp_delta.seconds) > \
                            tolerance_pct / 100 * exp_delta.seconds:
                        print("Warning: Timestamp of a data point of metric "
                              f"{metric_name!r} is not within expected delta "
                              "of its previous data point. "
                              f"Actual delta: {delta_td}, "
                              f"Expected delta: {exp_delta} "
                              f"(+/-{tolerance_pct}%)")

                previous_dt = dp_timestamp_dt


TESTCASES_PART_ATTACH_DETACH_NETWORK_LINK = [
    # Test cases for test_part_attach_detach_network_link(), each as a tuple
    # with these items:
    # * desc: Testcase short description
    # * type: Partition link type ("hipersockets", "smc-d")
    # * number_of_nics: Number of NICs to be created during attach
    # * nic_property_list: Properties of the NICs to be created during attach
    (
        "Hipersockets, 1 NIC with default properties",
        "hipersockets",
        1,
        []
    ),
    (
        "Hipersockets, 2 NICs with default properties",
        "hipersockets",
        2,
        []
    ),
    (
        "Hipersockets, 1 NIC with devno",
        "hipersockets",
        1,
        [
            {
                'device-number': '4711',
            },
        ]
    ),
    (
        "Hipersockets, 1 NIC with VLAN",
        "hipersockets",
        1,
        [
            {
                'vlan-id': 53,
            },
        ]
    ),
    (
        "SMC-D, 1 NIC with default properties",
        "smc-d",
        1,
        []
    ),
    (
        "SMC-D, 2 NICs with default properties",
        "smc-d",
        2,
        []
    ),
    (
        "SMC-D, 1 NIC with with devno",
        "smc-d",
        1,
        [
            {
                'device-number': '4711',
            },
        ]
    ),
    # Note: Disabled because it currently fails
    # (
    #     "SMC-D, 1 NIC with with FID",
    #     "smc-d",
    #     1,
    #     [
    #         {
    #             'fid': 53,
    #         },
    #     ]
    # ),
]


@pytest.mark.parametrize(
    "desc, type, number_of_nics, nic_property_list",
    TESTCASES_PART_ATTACH_DETACH_NETWORK_LINK)
def test_part_attach_detach_network_link(
        desc, type, number_of_nics, nic_property_list,
        dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument,redefined-builtin
    """
    Test for Partition.attach/detach_network_link()
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        console = cpc.manager.client.consoles.console
        session = cpc.manager.session
        hd = session.hmc_definition

        if hd.mock_file:
            skip_warn("zhmcclient mock does not support "
                      "Partition.attach/detach_network_link()")

        part = None
        partlink = None

        try:

            partlink_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
            partlink_input_props = {
                "name": partlink_name,
                "type": type,
                "cpc-uri": cpc.uri,
            }
            partlink = console.partition_links.create(partlink_input_props)
            part_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
            part = cpc.partitions.create(
                standard_partition_props(cpc, part_name))

            # The code to be tested
            part.attach_network_link(
                partlink, number_of_nics, nic_property_list)

            # Verify the partition link is now attached
            attached_pls = part.list_attached_partition_links()
            assert len(attached_pls) == 1
            attached_pl = attached_pls[0]
            assert attached_pl.uri == partlink.uri

            # Verify the number of NICs in the partition link data
            partlink.pull_full_properties()
            bc_list = partlink.get_property('bus-connections')
            assert len(bc_list) == 1   # = attached partitions
            bc = bc_list[0]
            assert bc["partition-uri"] == part.uri
            assert len(bc["nics"]) == number_of_nics

            # Verify the number of NIC element objects in the partition
            # Note: for SMC-D, NICs are not represented in the partition
            part.pull_full_properties()
            nics = part.nics.list()
            nic_uris = part.get_property('nic-uris')
            if type == "hipersockets":
                assert len(nics) == len(nic_uris) == number_of_nics
            else:  # type == "smc-d":
                assert len(nics) == len(nic_uris) == 0

            # The code to be tested
            part.detach_network_link(partlink)

            # Verify the partition link is no longer attached
            attached_pls = part.list_attached_partition_links()
            assert len(attached_pls) == 0

            # Verify the number of NICs in the partition link data
            partlink.pull_full_properties()
            bc_list = partlink.get_property('bus-connections')
            assert len(bc_list) == 0   # = attached partitions

            # Verify the number of NIC element objects in the partition
            # Note: for SMC-D, NICs are not represented in the partition
            part.pull_full_properties()
            nics = part.nics.list()
            nic_uris = part.get_property('nic-uris')
            if type == "hipersockets":
                # Note: Detaching a partition link currently does not delete
                # the NIC in the partition.
                if len(nics) > 0 or len(nic_uris) > 0:
                    # warnings.warn(
                    print("Debug: Detaching Hipersocket partition link did "
                          "not delete the NICs in the partition:\n"
                          f"{nic_uris}\n{nics}")
            else:  # type == "smc-d":
                assert len(nics) == len(nic_uris) == 0

        finally:
            if partlink:
                partlink.delete()
            if part:
                part.delete()


TESTCASES_PART_ATTACH_DETACH_CTC_LINK = [
    # Test cases for test_part_attach_detach_ctc_link(), each as a tuple
    # with these items:
    # * desc: Testcase short description
    # * num_partitions: Number of partitions for creating the PL
    # * num_paths: Number of paths for creating the PL
    (
        "2 partitions, 1 path",
        2,
        1,
    ),
    (
        "3 partitions, 2 paths",
        3,
        2,
    ),
]


@pytest.mark.parametrize(
    "desc, num_partitions, num_paths",
    TESTCASES_PART_ATTACH_DETACH_CTC_LINK)
def test_part_attach_detach_ctc_link(
        logger, dpm_mode_cpcs,  # noqa: F811
        desc, num_partitions, num_paths):
    # pylint: disable=redefined-outer-name,unused-argument,redefined-builtin
    """
    Test for Partition.attach/detach_ctc_link()
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        logger.debug("Testing with CPC %s", cpc.name)

        console = cpc.manager.client.consoles.console
        session = cpc.manager.session
        hd = session.hmc_definition

        if hd.mock_file:
            skip_warn("zhmcclient mock does not support "
                      "Partition.attach/detach_ctc_link()")

        pl_parts = []
        partlink = None
        test_part = None

        try:

            # Select the adapters for the paths of the partition link.
            # We use the same adapter for each connection, because if we use
            # different adapters, we cannot determine which ones actually do
            # have a physical connection.
            adapters = cpc.adapters.list(
                filter_args={'type': 'fc', 'status': 'active'})
            if len(adapters) < num_paths:
                pytest.skip(f"CPC {cpc.name} has less than {num_paths} "
                            "online FC adapters for CTC")
            path_adapters = random.sample(adapters, num_paths)

            # Create the initial partitions for the partition link
            for i in range(0, num_partitions):
                part_name = f"{TEST_PREFIX}_{i}_{uuid.uuid4().hex}"
                logger.debug("Preparing properties for creation of initial "
                             "PL partition %s", part_name)
                part_props = standard_partition_props(cpc, part_name)
                logger.debug("Creating initial PL partition %s", part_name)
                part = cpc.partitions.create(part_props)
                pl_parts.append(part)

            # Create the partition link
            partlink_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
            paths = []
            for i in range(0, num_paths):
                path = {
                    "adapter-port-uri": path_adapters[i].uri,
                    "connecting-adapter-port-uri": path_adapters[i].uri,
                }
                paths.append(path)
            partlink_input_props = {
                "name": partlink_name,
                "type": 'ctc',
                "cpc-uri": cpc.uri,
                "partitions": [p.uri for p in pl_parts],
                "paths": paths,
            }
            logger.debug("Creating partition link %s", partlink_name)
            partlink = console.partition_links.create(partlink_input_props)

            logger.debug("Partition link after creation (default properties):\n"
                         "%s", pformat_as_dict(partlink.properties))
            partlink.pull_full_properties()
            logger.debug("Partition link after creation (full properties):\n"
                         "%s", pformat_as_dict(partlink.properties))
            assert_ctc_partition_link(partlink, num_paths, num_partitions)

            # Create the test partition (to which the PL is attached)
            test_part_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
            logger.debug("Preparing properties for creation of "
                         "test partition %s", test_part_name)
            test_part_props = standard_partition_props(cpc, test_part_name)
            logger.debug("Creating test partition %s", test_part_name)
            test_part = cpc.partitions.create(test_part_props)

            # Verify the partition link is not attached.
            # The code to be tested
            attached_pls = test_part.list_attached_partition_links()
            assert len(attached_pls) == 0

            logger.debug("Attaching partition link to test partition")

            # The code to be tested
            test_part.attach_ctc_link(partlink)
            num_partitions += 1

            partlink.pull_full_properties()
            logger.debug("Partition link after attach (full properties):\n"
                         "%s", pformat_as_dict(partlink.properties))
            assert_ctc_partition_link(partlink, num_paths, num_partitions)

            # Verify the partition link is now attached.
            # The code to be tested
            attached_pls = test_part.list_attached_partition_links()
            assert len(attached_pls) == 1
            attached_pl = attached_pls[0]
            assert attached_pl.uri == partlink.uri

            logger.debug("Detaching partition link from test partition")

            # The code to be tested
            test_part.detach_ctc_link(partlink)
            num_partitions -= 1

            partlink.pull_full_properties()
            logger.debug("Partition link after detach (full properties):\n"
                         "%s", pformat_as_dict(partlink.properties))
            assert_ctc_partition_link(partlink, num_paths, num_partitions)

            # Verify the partition link is no longer attached.
            # The code to be tested
            attached_pls = test_part.list_attached_partition_links()
            assert len(attached_pls) == 0

        finally:
            if partlink:
                logger.debug("Deleting partition link %s", partlink.name)
                partlink.delete()
            if test_part:
                logger.debug("Deleting test partition %s", test_part.name)
                test_part.delete()
            for part in pl_parts:
                logger.debug("Deleting initial PL partition %s", part.name)
                part.delete()



# Copyright 2019,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Pytest fixtures for mocked HMCs.
"""


import pytest

from ._hmc_definition import HMCDefinition
from ._hmc_definitions import hmc_definitions
from ._hmc_session_functions import setup_hmc_session, teardown_hmc_session

__all__ = ['hmc_definition', 'hmc_session']


def fixtureid_hmc_definition(fixture_value):
    """
    Return a fixture ID to be used by pytest, for fixture `hmc_definition()`.

    Parameters:

      fixture_value (HMCDefinition): The HMC definition of the HMC the test
        runs against.
    """
    if not isinstance(fixture_value, HMCDefinition):
        return None  # Use pytest auto-generated ID
    hd = fixture_value
    show_hmc = False  # Enable to show HMC/userid or mock file in tests
    if not show_hmc:
        hmc_str = ""
    elif hd.mock_file:
        hmc_str = f"(mock_file={hd.mock_file})"
    else:
        hmc_str = f"(host={hd.host}, userid={hd.userid})"
    ret_str = f"hmc_definition={hd.nickname}{hmc_str}"
    return ret_str


@pytest.fixture(
    params=hmc_definitions(load=None),
    scope='module',
    ids=fixtureid_hmc_definition
)
def hmc_definition(request):
    """
    Pytest fixture representing the set of HMC definitions to use for a test.

    A test function parameter with the name of this fixture resolves to the
    :class:`~zhmcclient.testutils.HMCDefinition`
    object of each HMC to test against.

    The test function is called once for each HMC, if the targeted HMC is a
    group.
    """
    return request.param


@pytest.fixture(
    scope='module'
)
def hmc_session(request, hmc_definition):
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Pytest fixture representing the set of HMC sessions to run a test against.

    A test function parameter with the name of this fixture resolves to the
    :class:`zhmcclient.Session` or :class:`zhmcclient_mock.FakedSession` object
    for each HMC to test against.

    The session is already logged on to the HMC.

    The session object has an additional property named ``hmc_definition``
    that is the :class:`~zhmcclient.testutils.HMCDefinition` object for the
    corresponding HMC definition in the :ref:`HMC inventory file`.

    Because the `hmc_definition` parameter of this fixture is again a fixture,
    the :func:`zhmcclient.testutils.hmc_definition` function needs to be
    imported as well when this fixture is used.

    Upon fixture teardown, the session is automatically logged off from the HMC.
    """
    session = setup_hmc_session(hmc_definition)
    yield session
    teardown_hmc_session(session)



# Copyright 2025 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`MFA Server Definition` resource represents a definition that contains
information about an Multi-factor Authentication (MFA) server that may be used
for HMC user authentication purposes.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_MFA_SERVER_DEFINITION

__all__ = ['MfaServerDefinitionManager', 'MfaServerDefinition']


class MfaServerDefinitionManager(BaseManager):
    """
    Manager providing access to the :term:`MFA Server Definition` resources of
    a HMC.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Console` object:

    * :attr:`zhmcclient.Console.mfa_server_definitions`

    HMC/SE version requirements:

    * HMC version == 2.15.0
    """

    def __init__(self, console):
        # This function should not go into the docs.
        # Parameters:
        #   console (:class:`~zhmcclient.Console`):
        #      Console object representing the HMC.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        # Because this resource has case-insensitive names, this list must
        # contain the name property.
        query_props = [
            'name',
        ]

        super().__init__(
            resource_class=MfaServerDefinition,
            class_name=RC_MFA_SERVER_DEFINITION,
            session=console.manager.session,
            parent=console,
            base_uri='/api/console/mfa-server-definitions',
            oid_prop='element-id',
            uri_prop='element-uri',
            name_prop='name',
            query_props=query_props,
            case_insensitive_names=True)

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`: :term:`Console` defining the scope for
        this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the :term:`MFA Server Definition` resources representing the
        definitions of MFA servers in this HMC.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * HMC version == 2.15.0

        Authorization requirements:

        * User-related-access permission to the MFA Server Definition objects
          included in the result, or task permission to the "Manage Multi-factor
          Authentication" task.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.MfaServerDefinition` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.FilterConversionError`
        """
        result_prop = 'mfa-server-definitions'
        list_uri = f'{self.console.uri}/mfa-server-definitions'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)

    @logged_api_call(blanked_properties=['bind-password'], properties_pos=1)
    def create(self, properties):
        """
        Create a new MFA Server Definition in this HMC.

        HMC/SE version requirements:

        * HMC version == 2.15.0

        Authorization requirements:

        * Task permission to the "Manage Multi-factor Authentication" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create MFA Server Definition' in the :term:`HMC API`
            book.

        Returns:

          MfaServerDefinition:
            The resource object for the new MFA Server Definition.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(
            self.console.uri + '/mfa-server-definitions', body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        mfa_server_definition = MfaServerDefinition(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return mfa_server_definition


class MfaServerDefinition(BaseResource):
    """
    Representation of a :term:`MFA Server Definition`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.MfaServerDefinitionManager`).

    HMC/SE version requirements:

    * HMC version == 2.15.0
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.MfaServerDefinitionManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, MfaServerDefinitionManager), (
            "Console init: Expected manager type "
            f"{MfaServerDefinitionManager}, got {type(manager)}")
        super().__init__(
            manager, uri, name, properties)

    @logged_api_call
    def delete(self):
        """
        Delete this MFA Server Definition.

        HMC/SE version requirements:

        * HMC version == 2.15.0

        Authorization requirements:

        * Task permission to the "Manage Multi-factor Authentication" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call(blanked_properties=['bind-password'], properties_pos=1)
    def update_properties(self, properties):
        """
        Update writeable properties of this MFA Server Definitions.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * HMC version == 2.15.0

        Authorization requirements:

        * Task permission to the "MFA Server Definition Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'MFA Server Definition object' in
            the :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)

        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)
